#include <vector>
#include <stdlib.h> /* rand() */
#include <sys/time.h>
#include <set>
#include <map>
#include <sys/stat.h>

#include <boost/random/mersenne_twister.hpp>
#include <boost/random/uniform_int.hpp>
#include <boost/random/variate_generator.hpp>

#define DO_TRACK_PC 0
#define DO_DISASM   0
#define DO_RAMDUMP  1

#define BISQBOT_INTERNAL
#include "bisqbotdefs.hh"
#include "bisqbot.hh"
extern "C" unsigned char RAM[0x800];
extern "C" unsigned char PALRAM[0x20];
extern "C" void FCEU_DispMessage(const char *format, ...)
  __attribute__((format(printf,1,2)));

extern "C" unsigned char RefrRAM[0x800];
unsigned char RefrRAM[0x800];

static unsigned NextChange=0;

static boost::mt19937 rng;

#include "rockman1sound.hh"
static class Rockman1Sound Rockman1Sound;
#include "rockman1sound.cc"

template<unsigned mod>
static inline unsigned GetRnd()
{
    static boost::uniform_int<int> g(0, mod-1);
    return g(rng);
}
static inline unsigned GetRnd(unsigned mod)
{
    return boost::uniform_int<int> (0,mod-1) (rng);
}

struct BisqBotStateData
{
    /* Put here all variables that you want to be saved
     * with Savestates when the core does rerecording.
     * Reset() is only called when seeking back to frame 0,
     * which is not always the case if you use 2 or more steps.
     */
    int Goal;
    double BonusX, BonusY;
    unsigned NumShoots;
    unsigned MaxShoots;
    unsigned EventFrame;
    bool Region;
};
static class BisqBotState: public BisqBotStateBase, public BisqBotStateData
{
public:
    virtual ~BisqBotState() { }
    
    BisqBotState()
    {
        Reset();
    }
    void Reset()
    {
        Goal = 0;
        NextChange = 0;
        NumShoots = 0;
        BonusX = 0;
        BonusY = 0;
        MaxShoots = 0;
        Region = false;
        //EventFrame = 0;
        EventFrame = 0 + boost::uniform_int<int> (0,20) (rng);
    }
    
    void operator= (const BisqBotState& b)
    {
        (BisqBotStateData&)*this = b;
    }

public:
} state;

const BisqBotStateBase* BisqBotNewSaveState()
{
    return new BisqBotState(state);
}

void BisqBotLoadState(const BisqBotStateBase* b)
{
    state = *(BisqBotState*)b;
}

extern "C" {
void FCEUI_GetIVectors(unsigned short *reset, unsigned short *irq, unsigned short *nmi);
unsigned SavedHScroll,SavedVScroll;
#include "types.h"
extern readfunc ARead[0x10000];
}

static void Check_RAMdiffs(unsigned S)
{
    static unsigned prev_S;
    
    static std::map<unsigned/*sp*/,std::string> RAMstates;
    std::string cur_ramstate((char*)RAM, (char*)RAM+0x800);
    
    /* Ignore stack differences */
    for(unsigned a=0x180; a<=0x1FF; ++a) cur_ramstate[a] = 0;
    /* Ignore sprite table */
    for(unsigned a=0x200; a<=0x2FF; ++a) cur_ramstate[a] = 0;
    /* Ignore music player differences */
    for(unsigned a=0x500; a<=0x57D; ++a) cur_ramstate[a] = 0;
    for(unsigned a=0xE0; a<=0xF5; ++a) cur_ramstate[a] = 0;
    
    std::string& prev_ramstate = RAMstates[S];
    
    if(cur_ramstate != prev_ramstate)
    {
        if(S > prev_S) /* returning from function */
        {
            unsigned wid = 0;
            wid += printf("RAM diffs (to previous time when S=%02X): ", S);
            
            for(unsigned a=0; a<0x800
                              && a < prev_ramstate.size()
                              && a < cur_ramstate.size();
                              ++a)
            {
                if(cur_ramstate[a] != prev_ramstate[a])
                {
                    if(wid > 76) { printf("\n  "); wid = 2; }
                    wid += printf("%03X:%02X ", a, (unsigned char)cur_ramstate[a]);
                }
            }
            printf("\n");
        }
        
        prev_ramstate = cur_ramstate;
    }
    prev_S = S;
}

static const struct {const char*pat;unsigned bytes;bool rel;} addrmodes[12] =
{
    {"%.3s"         ,0,0},
    {"%.3s #%s"     ,1,0},
    {"%.3s %s"      ,1,1},
    {"%.3s %s"      ,1,0},
    {"%.3s %s,x"    ,1,0},
    {"%.3s %s,y"    ,1,0},
    {"%.3s (%s,x)"  ,1,0},
    {"%.3s (%s),y"  ,1,0},
    {"%.3s %s"      ,2,0},
    {"%.3s %s,x"    ,2,0},
    {"%.3s %s,y"    ,2,0},
    {"%.3s (%s)"    ,2,0},
};
static const char asminfo[] =
 // addressing modes
 "AGZZZDDZABAZZIIZ" "CHZZZEEZAKZZZJJZ" "IGZZDDDZABAZIIIZ" "CHZZZEEZAKZZZJJZ"
 "AGZZZDDZABAZIIIZ" "CHZZZEEZAKZZZJJZ" "AGZZZDDZABAZLIIZ" "CHZZZEEZAKZZZJJZ"
 "ZGZZDDDZAZAZIIIZ" "CHZZEEFZAKAZZJZZ" "BGBZDDDZABAZIIIZ" "CHZZEEFZAKAZJJKZ"
 "BGZZDDDZABAZIIIZ" "CHZZZEEZAKZZZJJZ" "BGZZDDDZABAZIIIZ" "CHZZZEEZAKZZZJJZ"
 // opcodes
 "brkora" "??????" "???ora" "asl???" "phpora" "asl???" "???ora" "asl???"
 "bplora" "??????" "???ora" "asl???" "clcora" "??????" "???ora" "asl???"
 "jsrand" "??????" "bitand" "rol???" "plpand" "rol???" "bitand" "rol???"
 "bmiand" "??????" "???and" "rol???" "secand" "??????" "???and" "rol???"
 "rtieor" "??????" "???eor" "lsr???" "phaeor" "lsr???" "jmpeor" "lsr???"
 "bvceor" "??????" "???eor" "lsr???" "clieor" "??????" "???eor" "lsr???"
 "rtsadc" "??????" "???adc" "ror???" "plaadc" "ror???" "jmpadc" "ror???"
 "bvsadc" "??????" "???adc" "ror???" "seiadc" "??????" "???adc" "ror???"
 "???sta" "??????" "stysta" "stx???" "dey???" "txa???" "stysta" "stx???"
 "bccsta" "??????" "stysta" "stx???" "tyasta" "txs???" "???sta" "??????"
 "ldylda" "ldx???" "ldylda" "ldx???" "taylda" "tax???" "ldylda" "ldx???"
 "bcslda" "??????" "ldylda" "ldx???" "clvlda" "tsx???" "ldylda" "ldx???"
 "cpycmp" "??????" "cpycmp" "dec???" "inycmp" "dex???" "cpycmp" "dec???"
 "bnecmp" "??????" "???cmp" "dec???" "cldcmp" "??????" "???cmp" "dec???"
 "cpxsbc" "??????" "cpxsbc" "inc???" "inxsbc" "nop???" "cpxsbc" "inc???"
 "beqsbc" "??????" "???sbc" "inc???" "sedsbc" "??????" "???sbc" "inc???";

extern "C" { extern unsigned char * ROM, *Page[32]; }

static void DisAsm(unsigned PC,unsigned A,unsigned X,unsigned Y,unsigned S,unsigned P)
{
    //Check_RAMdiffs(S);
    const unsigned AbsPC = (&Page[PC >> 11][PC]) - ROM;
    
    const unsigned char opcode = ARead[PC](PC);
    const unsigned addrmode = asminfo[opcode]-'A';
    const unsigned bytes = 1+addrmodes[addrmode].bytes;
    bool indexed = addrmode==9||addrmode==10;
    
    unsigned char RomPage = (AbsPC / 0x2000);
    printf("[%02X]%04X: ", RomPage, PC);
    
    for(unsigned b=0; b<bytes; ++b) printf("%02X ", ARead[PC+b](PC+b));
    for(unsigned b=bytes; b<5; ++b) printf("   ");
    char OpBuf[512];
    char OpParamStr[50];

    unsigned opparam = 0;
    for(unsigned b=1; b<bytes; ++b) opparam |= (ARead[PC+b](PC+b)) << (8*(b-1));
    if(addrmodes[addrmode].rel) opparam = PC + (signed char)opparam + 2;
    unsigned opbytes = bytes-1; if(addrmodes[addrmode].rel) ++opbytes;
    sprintf(OpParamStr, "$%0*X", opbytes*2, opparam);
    std::string o = OpParamStr;
    std::string oo = indexed ? "Object" : "Megaman";
    
    unsigned addrinfo = opparam;
    if(addrmode >= 3)
    {
        switch(opparam)
        {
#if 0 /* ROCKMAN 2 */
            case 0x400: o = oo+"Type"; break;
            case 0x420: o = oo+"Flags";  break;
            case 0x440: o = oo+"PosScreen"; break;
            case 0x460: o = oo+"PosX"; break;
            case 0x480: o = oo+"PosXFraction"; break;
            case 0x4A0: o = oo+"PosY"; break;
            case 0x4C0: o = oo+"PosYFraction"; break;
            case 0x600: o = oo+"SpeedX"; break;
            case 0x620: o = oo+"SpeedXFraction"; break;
            case 0x640: o = oo+"SpeedY"; break;
            case 0x660: o = oo+"SpeedYFraction"; break;
            case 0x6C0: o = oo+"LifeMeter"; break;
            case 0x200: o = "SpriteTable"; break;
            case 0x1D: o = "NMI_GfxUpdateDone"; break;
            case 0x1C: o = "FrameCounter"; break;
            case 0x14: o = "CurrentBeginScreen"; break;
            case 0x15: o = "CurrentEndScreen"; break;
            case 0x38: o = "CurrentOrderNum"; break;
            case 0x4A: o = "RandomSeed"; break;
            case 0x4B: o = "BlinkStatus"; break;
            case 0xB1: o = "BossStrategyNumber"; break;
            case 0xB2: o = "BossStrategyCounter"; break;
            case 0xB3: o = "BossNumber"; break;
            
            case 0xA8: o = "LifesRemaining"; break;
            case 0xC000: o = "SwitchBank"; break;
            case 0xC07F: o = "WaitNextFrame"; break;
            case 0xCC6C: o = "ClearSpriteTable"; break;
            case 0xCFA8: o = "RenderMeter"; break;
            case 0xCF5D: o = "RenderMeters"; break;
            
            case 0xC051: o = "PlaySFX"; break;
            
            case 0xCEF9: o = "RenderSomething"; break;
            case 0xCC77: o = "RenderEverything"; break;
            
            case 0xD0D7: o = "ReadJoypads"; break;
            
            case 0xEEEF: o = "MoveObject"; break;
#endif
#if 0 /* LEGEND OF KAGE */
            case 0x0106: o = "PPU2000"; break;
            case 0x0107: o = "PPU2001"; break;
            case 0x0109: o = "PPUXScroll"; break;
            case 0x010A: o = "PPUYScroll"; break;
            case 0x010D: o = "PPUWriteBuf_32bytes"; break;
            case 0xA48E: o = "UpdateScrolling"; break;
            case 0xA7BC: o = "UpdateScrolling2"; break;
            case 0xA4E5: o = "UpdateRandom"; break;
            case 0x0F: o = "RndSeed"; break;
            case 0xC6: o = "NMIactiveflag"; break;
            case 0x0101: o = "KeyPadThisFrame"; break;
            case 0x0102: o = "KeyPadPrevFrame"; break;
            case 0x0103: o = "KeyPadThisFrameCopy"; break;
            case 0x0105: o = "NewKeysThisFrame"; break;
            case 0xC7: o = "GamePaused"; break;
            case 0x6A2: o = "Channel0SFXptr_lo"; break;
            case 0x6A3: o = "Channel0SFXptr_lo"; break;
            case 0x6B2: o = "Channel1SFXptr_lo"; break;
            case 0x6B3: o = "Channel1SFXptr_lo"; break;
            case 0x6C2: o = "Channel2SFXptr_lo"; break;
            case 0x6C3: o = "Channel2SFXptr_lo"; break;
            case 0x6D2: o = "Channel3SFXptr_lo"; break;
            case 0x6D3: o = "Channel3SFXptr_lo"; break;
            case 0x6E2: o = "Channel4SFXptr_lo"; break;
            case 0x6E3: o = "Channel4SFXptr_lo"; break;
#endif
#if 0 /* Castlevania */
            case 0x18: o="GameMode";break;
            case 0x19: o="GameModeAux";break;
            case 0x1A: o="FrameCounter";break;
            case 0x24: o="BankNumber6";break;
            case 0x27: o="CurrentBank";break;
            case 0x28: o="CurrentStage";break;
            case 0x2E: o="NumberofLives";break;
            case 0x3C: o="SimonMovingDirection";break;
            case 0x3E: o="SimonControllableFlag";break;
            case 0x3F: o="SimonYcoord";break;
            case 0x40: o="SimonXcoordLo";break;
            case 0x41: o="SimonXcoordHi";break;
            case 0x42: o="TimeLeftLo";break;
            case 0x43: o="TimeLeftHi";break;
            case 0x44: o="SimonDisplayedHP";break;
            case 0x45: o="SimonActualHP";break;
            case 0x64: o="CurrentMultiplier";break;
            case 0x6F: o="CurrentRandomBonusId";break;
            case 0x70: o="WhipLength012";break;
            case 0x71: o="NumberofHearts";break;
            case 0x72: o="SpecialWeaponOnScreen";break;
            case 0x79: o="MultiplierSpawnCounter";break;
            case 0x7B: o="SpecialItemCounter";break;
            case 0x7F: o="InSoundCode";break;
            case 0x015B: o="CurrentWeapon";break;
            case 0x1A9: o="Boss1A9";break;
            case 0x1AA: o="Boss1AA";break;
            case 0x58B: o="Mummy1";break;
            case 0x58C: o="Mummy2";break;
            case 0x300: o="ObjectUnknown300";break;
            case 0x31C: o="ObjectPose";break;
            case 0x338: o="ObjectPaletteIndex";break;
            case 0x354: o="ObjectYcoord";break;
            case 0x370: o="ObjectUnknown354";break;
            case 0x38C: o="ObjectScreenXcoordInt";break;
            case 0x3A8: o="ObjectScreenXcoordFrac";break;
            case 0x3C4: o="ObjectUnknown3C4";break;
            case 0x3E0: o="ObjectUnknown3E0";break;
            case 0x3FC: o="ObjectXSpeedInt";break;
            case 0x418: o="ObjectXSpeedFrac";break;
            case 0x434: o="ObjectUnknown434";break;
            case 0x450: o="ObjectUnknown450";break;
            case 0x46C: o="ObjectCurrentActionType";break;
            case 0x488: o="ObjectJumpLengthCounter";break;
            case 0x4A4: o="ObjectPoseChangeCounter";break;
            case 0x4C0: o="ObjectPoseTableIndex";break;
            case 0x4DC: o="ObjectJumpVar1";break;
            case 0x4F8: o="ObjectJumpVar2";break;
            case 0x514: o="ObjectMovingDownwardsFlag";break;
            case 0x530: o="ObjectScreenXcoordFrac";break;
            case 0x54C: o="ObjectUnknown54C";break;
            case 0x568: o="ObjectMultipurposeCounter";break;
            case 0x584: o="ObjectBonusType";break;
#endif
#if 0 /* Lunar ball */
            case 0xC000: o="CodeC000";break;
            case 0xCB5E: o="IncZeroPageWordAtX";break;
            case 0xCB57: o="IncWordAt00FA";break;
            case 0xFDDF: o="NegWordAt00FC";break;
            case 0xCB8F: o="SwapWordsAt00FAand00FC";break;
            case 0xCBB6: o="OutputStringTo2007from00FA";break;
            case 0xCBD9: o="DoSomethingString";break;
            case 0xCBC6: o="OutputText";break;
            case 0xCF98: o="DoSomethingFifths";break;
            case 0xEBAE: o="Text_Stage";break;
            case 0xEBC2: o="Text_Start";break;
            case 0xF4DC: o="Text_GameOver";break;
            case 0xEEAD: o="BallAI_AlreadyPocketed";break;
            case 0xEFB2: o="Ball_BeginVanishing";break;
            case 0xEFD4: o="VanishAllBalls";break;
            case 0xFE63: o="Sin";break;
            case 0xFEC9: o="Cos";break;
            case 0x08: o="NumBalls";break;
            case 0x1C8: o="NMIenteredFlag";break;
            case 0xCC8C: o="PPUDoWriteAndRead";break;
            case 0xCC78: o="PPUResetWriteAndRead";break;
            case 0xCCE2: o="PPUSendPalettes";break;
            case 0xCD0B: o="PPUReadJoypads";break;
            case 0xC180: o="Joy1data";break;
            case 0xC182: o="Joy1connected";break;
            case 0xC181: o="Joy2data";break;
            case 0xC183: o="Joy2connected";break;
            case 0x80 : o="PPUImagePaletteArray";break;
            case 0x90 : o="PPUSpritePaletteArray";break;
            case 0xD0 : o="PPUWriteCount";break;
            case 0xD1 : o="PPUWriteDataIndex";break;
            case 0xD2 : o="PPUWriteAddressLo";break;
            case 0xD3 : o="PPUWriteAddressHi";break;
            case 0xA0 : o="PPUWriteData";break;
            case 0xED : o="PPUWriteIndex";break;
            case 0xEE : o="PPUWriteDataUsedLength";break;
            case 0xE8 : o="PPUReadCount";break;
            case 0xE9 : o="PPUReadAddressLo";break;
            case 0xEA : o="PPUReadAddressHi";break;
            case 0xEB : o="PPUReadTargetLo";break;
            case 0xEC : o="PPUReadTargetHi";break;
            case 0xF0 : o="PPU2000reg";break;
            case 0xF1 : o="PPU2001reg";break;
            case 0xCA35: o="PPUDisableVblankInterrupt";break;
            case 0xCA45: o="PPUEnableVblankInterrupt";break;
            case 0xCA2B: o="PPUEnableSpritesAndScreen";break;
            case 0xCA1E: o="PPUDisableSpritesAndScreen";break;
            case 0xCA4F: o="PPUResetScrolling";break;
            case 0xCB1A: o="PPUSet2000and2001";break;
            case 0xCB22: o="PPUSendSpriteTable";break;
            case 0x200: o="SpriteTable";break;
            case 0x300: o="CueStatus";break;
            case 0x301: o="Ball1Status";break;
            case 0x302: o="Ball2Status";break;
            case 0x303: o="Ball3Status";break;
            case 0x304: o="Ball4Status";break;
            case 0x305: o="Ball5Status";break;
            case 0x306: o="Ball6Status";break;
            case 0x310: o="CueIdentity";break;
            case 0x311: o="Ball1Identity";break;
            case 0x312: o="Ball2Identity";break;
            case 0x313: o="Ball3Identity";break;
            case 0x314: o="Ball4Identity";break;
            case 0x315: o="Ball5Identity";break;
            case 0x316: o="Ball6Identity";break;
            case 0x328: o="AimY";break;
            case 0x320: o="CueXLo";break;
            case 0x321: o="Ball1XLo";break;
            case 0x322: o="Ball2XLo";break;
            case 0x323: o="Ball3XLo";break;
            case 0x324: o="Ball4XLo";break;
            case 0x325: o="Ball5XLo";break;
            case 0x326: o="Ball6XLo";break;
            case 0x330: o="CueXHi";break;
            case 0x331: o="Ball1XHi";break;
            case 0x332: o="Ball2XHi";break;
            case 0x333: o="Ball3XHi";break;
            case 0x334: o="Ball4XHi";break;
            case 0x335: o="Ball5XHi";break;
            case 0x336: o="Ball6XHi";break;
            case 0x338: o="AimX";break;
            case 0x340: o="CueXdeltaLo";break;
            case 0x341: o="Ball1XdeltaLo";break;
            case 0x342: o="Ball2XdeltaLo";break;
            case 0x343: o="Ball3XdeltaLo";break;
            case 0x344: o="Ball4XdeltaLo";break;
            case 0x345: o="Ball5XdeltaLo";break;
            case 0x346: o="Ball6XdeltaLo";break;
            case 0x348: o="AimingDelayCounter";break;
            case 0x350: o="CueXdeltaHi";break;
            case 0x351: o="Ball1XdeltaHi";break;
            case 0x352: o="Ball2XdeltaHi";break;
            case 0x353: o="Ball3XdeltaHi";break;
            case 0x354: o="Ball4XdeltaHi";break;
            case 0x355: o="Ball5XdeltaHi";break;
            case 0x356: o="Ball6XdeltaHi";break;
            case 0x360: o="CueYLo";break;
            case 0x361: o="Ball1YLo";break;
            case 0x362: o="Ball2YLo";break;
            case 0x363: o="Ball3YLo";break;
            case 0x364: o="Ball4YLo";break;
            case 0x365: o="Ball5YLo";break;
            case 0x366: o="Ball6YLo";break;
            case 0x370: o="CueYHi";break;
            case 0x371: o="Ball1YHi";break;
            case 0x372: o="Ball2YHi";break;
            case 0x373: o="Ball3YHi";break;
            case 0x374: o="Ball4YHi";break;
            case 0x375: o="Ball5YHi";break;
            case 0x376: o="Ball6YHi";break;
            case 0x390: o="CueXspeedHi";break;
            case 0x391: o="Ball1XspeedHi";break;
            case 0x392: o="Ball2XspeedHi";break;
            case 0x393: o="Ball3XspeedHi";break;
            case 0x394: o="Ball4XspeedHi";break;
            case 0x395: o="Ball5XspeedHi";break;
            case 0x396: o="Ball6XspeedHi";break;
            case 0x380: o="CueXspeedLo";break;
            case 0x381: o="Ball1XspeedLo";break;
            case 0x382: o="Ball2XspeedLo";break;
            case 0x383: o="Ball3XspeedLo";break;
            case 0x384: o="Ball4XspeedLo";break;
            case 0x385: o="Ball5XspeedLo";break;
            case 0x386: o="Ball6XspeedLo";break;
            case 0x3A0: o="CueVelocityHi";break;
            case 0x3A1: o="Ball1VelocityHi";break;
            case 0x3A2: o="Ball2VelocityHi";break;
            case 0x3A3: o="Ball3VelocityHi";break;
            case 0x3A4: o="Ball4VelocityHi";break;
            case 0x3A5: o="Ball5VelocityHi";break;
            case 0x3A6: o="Ball6VelocityHi";break;
            case 0x3B0: o="CueVelocityLo";break;
            case 0x3B1: o="Ball1VelocityLo";break;
            case 0x3B2: o="Ball2VelocityLo";break;
            case 0x3B3: o="Ball3VelocityLo";break;
            case 0x3B4: o="Ball4VelocityLo";break;
            case 0x3B5: o="Ball5VelocityLo";break;
            case 0x3B6: o="Ball6VelocityLo";break;
            case 0x3C0: o="AimAngle";break;
            case 0x3D0: o="CueBallStatus";break;
            case 0xC003: o="NMISoundUpdater";break;
            case 0x68  : o="SoundEnabled";break;
            case 0xC5E0: o="SoundRegister4000value";break;
            case 0xC5E1: o="SoundRegister4001value";break;
            case 0xC5E2: o="SoundRegister4002value";break;
            case 0xC5E3: o="SoundRegister4003value";break;
            case 0xC5E4: o="SoundRegister4004value";break;
            case 0xC5E5: o="SoundRegister4005value";break;
            case 0xC5E6: o="SoundRegister4006value";break;
            case 0xC5E7: o="SoundRegister4007value";break;
            case 0xC5E8: o="SoundRegister4008value";break;
            case 0xC5E9: o="SoundRegister4009value";break;
            case 0xC5EA: o="SoundRegister400Avalue";break;
            case 0xC5EB: o="SoundRegister400Bvalue";break;
            case 0xC5EC: o="SoundRegister400Cvalue";break;
            case 0xC5ED: o="SoundRegister400Dvalue";break;
            case 0xC5EE: o="SoundRegister400Evalue";break;
            case 0xC5EF: o="SoundRegister400Fvalue";break;
            case 0xC5F0: o="SoundRegister4015value";break;
            case 0x70: o="SoundRegister4000changed";break;
            case 0x71: o="SoundRegister4001changed";break;
            case 0x72: o="SoundRegister4002changed";break;
            case 0x73: o="SoundRegister4003changed";break;
            case 0x74: o="SoundRegister4004changed";break;
            case 0x75: o="SoundRegister4005changed";break;
            case 0x76: o="SoundRegister4006changed";break;
            case 0x77: o="SoundRegister4007changed";break;
            case 0x78: o="SoundRegister4008changed";break;
            case 0x79: o="SoundRegister4009changed";break;
            case 0x7A: o="SoundRegister400Achanged";break;
            case 0x7B: o="SoundRegister400Bchanged";break;
            case 0x7C: o="SoundRegister400Cchanged";break;
            case 0x7D: o="SoundRegister400Dchanged";break;
            case 0x7E: o="SoundRegister400Echanged";break;
            case 0x7F: o="SoundRegister400Fchanged";break;
            case 0xCF88: o="UnreferencedCF88";break;
            case 0xCA8C: o="UnreferencedCA8C";break;
            case 0xCA94: o="UnreferencedCA94";break;
            case 0xCA9B: o="UnreferencedCA9B";break;
            case 0xCAC6: o="UnreferencedCAC6";break;
            case 0xCABE: o="UnreferencedCABE";break;
            case 0xCACC: o="UnreferencedCACC";break;
            case 0xCB3B: o="UnreferencedCB3B";break;
            case 0xCB40: o="UnreferencedCB40";break;
            case 0xCB52: o="UnreferencedCB52";break;
            case 0xCC63: o="UnreferencedCC63";break;
            case 0xCD35: o="UnreferencedCD35";break;
            case 0xE1CF: o="UnreferencedE1CF";break;
            case 0xF295: o="UnreferencedF295";break;
            case 0xFD3D: o="Unreferenced3D3D";break;
            
         // perl -pe 's/\$0([0-9]{3})/\$C$1/g'
         // perl -pe 's/\$1([0-9]{3})/\$D$1/g'
         // perl -pe 's/\$2([0-9]{3})/\$E$1/g'
         // perl -pe 's/\$3([0-9]{3})/\$F$1/g'
         // egrep '^(CertainlyCode|CertainlyData|RAM)'
         // sed 's/.*\$//'
         // perl -pe 's/(.*) (.*)/case 0x$1: o="$2";break;/'
#endif

        }
        if(addrmode == 4 || addrmode == 9) opparam += X;
        if(addrmode == 5 || addrmode == 10) opparam += Y;
    }
    sprintf(OpBuf, addrmodes[addrmode].pat, &asminfo[256+opcode*3], o.c_str());
    
    o = OpBuf;
    for(unsigned tmp = 0xFF; tmp > S; --tmp) o = " " + o;
    
    char OpInfo[512] = "";
    if(addrmode >= 3 && addrmode != 6 && addrmode != 7
    && addrmode != 11)
    {
        sprintf(OpInfo, " [$%04X]", opparam);
    }
    printf("%-30s %10s A=%02X X=%02X Y=%02X S=%02X P=%02X\n", o.c_str(), OpInfo, A,X,Y, S, P);
}

void BisqBotMonitorCPU
  (unsigned short PC,
   unsigned char A,
   unsigned char X,
   unsigned char Y,
   unsigned char S,
   unsigned char P)
{
    static unsigned short irq=0, nmi=0, res=0;
    if(!irq) { FCEUI_GetIVectors(&res,&irq,&nmi); }

    static unsigned prevpc=0;
    { static unsigned curpc; prevpc=curpc; curpc=PC; }
    
    const unsigned AbsPC = (&Page[PC >> 11][PC]) - ROM;

#if DO_TRACK_PC
   #if 0
    static std::map<unsigned, std::set<unsigned> > jptpc;
    if(/*prevpc == 0x83EE
    || prevpc == 0x8276
    || prevpc == 0x8803
    || prevpc == 0x8EBD
    || prevpc == 0xB821
    || prevpc == 0xF243
      */ARead[prevpc](prevpc) == 0x20) // jsr
    {
        if(jptpc[prevpc].find(PC) == jptpc[prevpc].end())
        {
            FILE*fp = fopen("bisqbot_pc.log", "a");
            if(fp)
                fprintf(fp,
                    "New PC $%X (ROMPTR $%X) called from $%X\n", PC, AbsPC, prevpc);
            fclose(fp);
            jptpc[prevpc].insert(PC);
        }
    }
   #endif
   #if 0 /* Solomon's Key */
    if(PC == 0xC563) { state.Goal += 1; FCEU_DispMessage("Got KEY\n"); }
    if(PC == 0xC587) { state.Goal += 2; FCEU_DispMessage("Got DOOR\n"); }
   #endif
#endif

#if 0 /* CASTLEVANIA */
    unsigned RockY  = RAM[0x3F];
    unsigned RockX  = RAM[0x41]*256 + RAM[0x40];
  //  if((PC|0xC000) == 0xEB81) { FCEU_DispMessage("ObjType[%u]=%u\n", X,A); }
    if(PC == 0xE65D) { FCEU_DispMessage("Multiplier incremented\n", X,A); }
    if(false && (PC|0xC000) == 0xE75B)
    {
        FCEU_DispMessage("HP set %u (was %u), X=%u & Y=%u\n", A, RAM[0x45], RockX, RockY);
        if(RockY > 152 || RockX >= 72) state.Goal = -1;
        else
            { FCEU_DispMessage("Possible success");
              BisqBotDisplaySeqSample(); 
            }
    }
    /*
    static long MiX=99999999,MaX=-99999,MiY=99999999,MaY=-999999;
    if(RockX < 55 && RockY <= 140) state.Goal += 2;
    if(RockY >= 192 || RockX >= 166) state.Goal = -1;
    if(RockX < MiX) { FCEU_DispMessage("New smallest X: %ld", MiX=RockX); BisqBotDisplaySeqSample(); }
    if(RockY < MiY) { FCEU_DispMessage("New smallest Y: %ld", MiY=RockY); BisqBotDisplaySeqSample(); }
    if(RockX > MaX) { FCEU_DispMessage("New biggest X: %ld", MaX=RockX); BisqBotDisplaySeqSample(); }
    if(RockY > MaY) { FCEU_DispMessage("New biggest Y: %ld", MaY=RockY); BisqBotDisplaySeqSample(); }
    */
    if(PC == 0xED34) FCEU_DispMessage("Boss DELAY var set %d (bonus id %02X, frame %02X)", A, RAM[0x6F], RAM[0x1A]);
#endif

#if 0
unsigned addr=PC;
    if(0
|| (addr >= 0x80F0 && addr <= 0x80F2)
|| (addr >= 0x81D8 && addr <= 0x81D8)
|| (addr >= 0x81DA && addr <= 0x81DA)
)
    {
        BotFrontMsg("PC=%X A=%02X X=%02X Y=%02X prevPC=%04X reset=%04X irq=%04X nmi=%04X\n",
            PC,A,X,Y, prevpc, res,irq,nmi);
       // exit(1);
    }
#endif

#if DO_DISASM /* DISASSEMBLY */
  //if(PC >= 0xB425 && PC <= 0xB465)
  {
    static int InSoundCode=0;
    static unsigned UnLoopPtr=0;
    static bool Looping=false;
 #if 1 /* disable for castlevania */
    if(PC == irq) { printf("IRQ\n"); InSoundCode=0; Looping=false; }
    if(PC == nmi) { printf("NMI\n"); InSoundCode=0; Looping=false; }
    if(PC == res) { printf("RESET\n"); InSoundCode=0; Looping=false; }
 #endif
    if(PC == UnLoopPtr) { /*printf("Loop break\n");*/ Looping=false; }
   #if 0 /* ROCKMAN 2 */
    if(PC == 0xD0AD) { InSoundCode=0; /*printf("exit sound code\n");*/ }
    if(PC == 0xD0BD) { InSoundCode=0; /*printf("exit sound code\n");*/ }
    if(PC == 0xC000) InSoundCode++; //switchbank
    if(PC == 0xD0D7) InSoundCode++; //readjoypads
    if(PC == 0xCFA8) InSoundCode++; //RenderMeter
    if(PC == 0xCF5D) InSoundCode++; //RenderMeters
    if(PC == 0xCEF9) InSoundCode++; //RenderSomething
    if(PC == 0xCC77) InSoundCode++; //RenderEverything
   #endif
   #if 0 /* KAGE */
   // if(PC == 0x8AA5) { InSoundCode=0; /* exit sound code */ }
    if(PC == 0xA48E) InSoundCode++; //horizscroll
    if(PC == 0xA7BC) InSoundCode++; //updategfx
    if(PC == 0xA4E5) InSoundCode++; //updaterandom
   #endif
   #if 0 /* CASTLEVANIA */
    //if(PC == 0xC052) InSoundCode += 100; // NMI has nothing important
    //if(PC == 0xC0BE) InSoundCode -= 100; // exit NMI
    //if(PC == 0x838A) InSoundCode++; //soundcode
    //if(PC == 0xA7BC) InSoundCode++; //updategfx
    //if(PC == 0xA4E5) InSoundCode++; //updaterandom
    //if(PC == 0xC09A) --InSoundCode; //exit sound code
    if(PC == 0xE931) { InSoundCode = 0; // object processing
                       printf("*** OBJ %02X type %02X\n", X,A);
                       if(A != 0x04 && A != 0x1D) InSoundCode = 1; // ignore
                     }
    if(PC == 0xE92B) InSoundCode = 1;
    if(PC == nmi) ++InSoundCode;
    if(PC == irq) --InSoundCode;
    if(PC >= 0xC030 && PC <= 0xC03C) InSoundCode=1;
    if(PC == 0xCA6D) ++InSoundCode;
    if(PC == 0xCA88) --InSoundCode;
    
   #endif
   
    if(!InSoundCode && !Looping)
    {
        DisAsm(PC,A,X,Y,S,P);
        if((ARead[PC](PC) & 0x1F) == 0x10 && (ARead[PC+1](PC+1) >= 0xE0))
        {
            unsigned jump_target = PC + (signed char)(ARead[PC+1](PC+1)) + 2;
            bool valid_short_jump = true;
            while(jump_target < PC)
            {
                unsigned char opcode = ARead[jump_target](jump_target);
                const unsigned addrmode = asminfo[opcode]-'A';
                const unsigned bytes = 1+addrmodes[addrmode].bytes;
                /* Don't hide this short loop, if it contains a branch or a call */
                if(opcode == 0x20
                || ((opcode&0x1F)==0x10)
                || opcode==0x4C
                || opcode==0x6C
                || opcode==0x60)
                {
                    valid_short_jump=false;
                    break;
                }
                jump_target += bytes;
            }
            
            if(valid_short_jump)
            {
                Looping=true;
                UnLoopPtr = PC+2;
                printf("<Short loop suppressed>\n");
            }
        }
    }
   #if 0 /* ROCKMAN 2 */
    if(PC == 0xC021) { printf("<function suppressed>\n"); InSoundCode--; } //switchbank
    if(PC == 0xD0F4) { printf("<function suppressed>\n"); InSoundCode--; } //readjoypads
    if(PC == 0xCFE2) { printf("<function suppressed>\n"); InSoundCode--; } //RenderMeter
    if(PC == 0xCFA7) { printf("<function suppressed>\n"); InSoundCode--; } //RenderMeters
    if(PC == 0xCEF6) { printf("<function suppressed>\n"); InSoundCode--; } //RenderSomething
    if(PC == 0xCD04) { printf("<function suppressed>\n"); InSoundCode--; } //RenderEverything
    if(PC == 0xD0AA) { InSoundCode++; printf("<sound code call suppressed>\n"); }
    if(PC == 0xD0BA) { InSoundCode++; printf("<sound code call suppressed>\n"); }
   #endif
   #if 0 /* KAGE */
   // if(PC == 0x8AA2) { InSoundCode++; printf("<sound code call suppressed>\n"); }
    if(PC == 0xA4DB) { printf("<function suppressed>\n"); InSoundCode--; } //RenderMeter
    if(PC == 0xA879) { printf("<function suppressed>\n"); InSoundCode--; } //updategfx
    if(PC == 0xA4F3) { printf("<function suppressed>\n"); InSoundCode--; } //updaterandom
   #endif
    if(InSoundCode<0)InSoundCode=0;
    fflush(stdout);
    
    /* Randomness in Kage does not happen in NMI. Routine is A4E5
          tmpvar is at [$0F]
          tmpvar = tmpvar * 5 + 17
     */
  }
#endif

#if 0 /* CV2 */
    static int BestYdiff = -100;
    int ydiff = RAM[0x324] - RAM[0x32B];
    if(ydiff > 3) state.Goal = 2;
    if(ydiff > BestYdiff)
    {
        fprintf(stderr, "Best Y: %02X (gypsy %02X), diff=%d\n", RAM[0x324], RAM[0x32B], ydiff);
        BestYdiff = ydiff;
    }
#endif
#if 0 /* SOLOKEY */
    if(PC == 0xC4D0)
    {
        extern int BqtItemAcquired;
        BqtItemAcquired = A;
    }
    if(PC == 0xC4B6)
    {
        if(!BotFrontActive())
            BotFrontMsg("Item %02X acquired (map tile collision)\n", A-6); 
    }
    if(PC == 0xA56C)
    {
        BotFrontMsg("Item %02X acquired from enemy\n", A); 
        extern int BqtItemAcquired;
        BqtItemAcquired = A+0x80;
        
        /* 01 = ?? 1up?
         * 02 = blue bottle, converts a regular fireball into a long one
         * 03 = scroll extension
         * 04 = yellow jewel (no points, affects fireballs)
         * 05 = fairy bell
         * 06 = red bottle labelled "E" (1up)
         * 07 = ?? 1up?
         * 08 = gray I-sac         (100)
         * 09 = gray II-sac
         * 0B = red I-sac          (1000)
         * 0C = red II-sac
         * 0D = red V-sac
         * 0E = grn I-sac          (10000)
         * 0F = grn II-sac
         * 10 (not exist) = 50000, 11 (not exist) = 100000, ...
         */
    }
#endif
    
#if 0 //ROCKMAN 1
    const unsigned bank = AbsPC / 0x4000;
    
    static int Prev_EnemyType[32] = {0};
    for(unsigned a=0; a<32; ++a)
    {
        unsigned type=0;
        if(RAM[0x600+a] >= 0xF8) type=0;
        else type=RAM[0x6E0+a];
        if(type != Prev_EnemyType[a])
        {
          //  fprintf(stderr, "Object %u is now %u ($%02X)\n", a, type, type);
            Prev_EnemyType[a] = type;
        }
    }
    
    
    //if(sound && PC == 0x9000) Rockman1Sound.Update();
    //if(sound && PC == 0x9003) Rockman1Sound.Command(A,Y);
    
    if(bank == 5 && PC == 0x91CF) { 
        //fprintf(stderr, "Sel registered\n");
        extern int BqtSelPushed; BqtSelPushed=1; }
    if(bank == 5 && PC == 0xA71E) { extern int BqtFiring; BqtFiring=1; }
    
    if(bank == 5 && PC == 0xA238)
    {
        FCEU_DispMessage("Damage from enemy %02X (%u points)\n", Y,A);
        extern int BqtDamaged;
        BqtDamaged = 1;
        state.Goal = -5;
    }
    
    if(bank == 5 && PC == 0xAD11)
    {
        FCEU_DispMessage("Suzy %u delay set as %02X\n", X, A);
    }
    
    if(bank == 5 && PC == 0xBEDF)
    {
        if(!BotFrontActive())
        FCEU_DispMessage("Enemy %02X hit with weapon %u (%d points left) - rndseed=%02X\n",
            (unsigned)Y, (unsigned)RAM[0x5F], (signed char)A, (unsigned)RAM[0x46]);
        /*if(A==0)
            switch(RAM[0x46])
            {
                case 0x5F: case 0x60:
                case 0xC3: case 0xC4:
                    break;
                default:
                    state.Goal = -5;
            }*/
        if(Y == 0x3A) state.Goal = 2; // bubblebot
    }
    if(bank == 5
    && (PC == 0x9FE8)) // horiz scrolling
    {
        FCEU_DispMessage("Scrolling routine at PC=%04X A=%02X X=%02X Y=%02X\n", PC, A,X,Y);
        state.Goal = 2;
    }
    if(bank == 5
    && (PC == 0x925D)) // scroll prev room
    {
        FCEU_DispMessage("Scroll previous room at PC=%04X A=%02X X=%02X Y=%02X\n", PC, A,X,Y);
    }
    if(bank == 5
    && (PC == 0x92A6)) // scroll next room
    {
        FCEU_DispMessage("Scroll next room at PC=%04X A=%02X X=%02X Y=%02X\n", PC, A,X,Y);
        state.Goal = 2;
        extern int BqtScrolled; BqtScrolled=1;
    }
    if(PC == 0xC912)
    {
        extern int BqtBossHit, BqtBossRemHp; BqtBossHit=1; BqtBossRemHp=A; 
        //state.Goal += 2;
        FCEU_DispMessage("Boss %u damaged with weapon %u (%d points left)\n", RAM[0xAC], RAM[0x5F], (signed char)A);
    }
    if(PC == 0xC833)
    {
        FCEU_DispMessage("Acquire item 0x%02X\n", A);
        
        //if(RAM[0x46] >= 0x30) state.Goal += 2; else
        //if(A == 0x3F) state.Goal += 2; else
        state.Goal = -5;
    }
    if(PC == 0xCE53 || PC == 0xCE2C)
    {
        //state.Goal += 1;
        FCEU_DispMessage("Rendered a demolished Gutsblock (Goal=%d)\n", (int)state.Goal);
        //state.EventFrame = 3;
    }
    if(bank == 5
    && (PC == 0xBF25))
    {
        unsigned objnum = RAM[0x2F];
        state.BonusX = RAM[0x460+objnum]*256 + (unsigned short)(RAM[0x480+objnum]*256 + RAM[0x4A0+objnum])/ 256.0;
        state.BonusY = (unsigned short)(RAM[0x600+objnum]*256 + RAM[0x620+objnum])/ 256.0;

        if(Y == 0x3F) state.Goal += 2; else state.Goal = -5;

        void (*method)(const char*, ...) =
            Y == 0x3F ? BotFrontMsg : FCEU_DispMessage;
        

        if(Y == 0x3B)
            method("Create no bonus at %.3f,%.3f\n", state.BonusX,state.BonusY);
        else
            method("Create bonus 0x%02X at %.3f,%.3f\n", (unsigned)Y, state.BonusX,state.BonusY);

        /* // wily2 xcoordinates
        if(state.BonusX < 6867) state.Goal = -5;
        if(state.BonusX > 6941.0) state.Goal = -5;
        if(state.BonusX > 6885.0 && state.BonusX < 6905.0) state.Goal = -5;
        */
        //if(state.BonusX < 6872.0) state.Goal = -5;
    }
    
    if(PC == 0xD4B1 && !BotFrontActive()) FCEU_DispMessage("LAG detected\n");
    if(PC == 0xE0E6) fprintf(stderr, "You're too close to Bombman\n");
    
    if(PC == 0xC219) { extern int BqtMegaKilled; BqtMegaKilled=1; }
#endif
    if(PC == 0xD4AD) std::memcpy(RefrRAM, RAM, 0x800);

#if 0  /* KAGE */
    if(PC == 0x973F) BotFrontMsg("3000 points at 973F\n");
    if(PC == 0x9765) BotFrontMsg("3000 points at 9765\n");
    if(PC == 0x9854) BotFrontMsg("3000 points at 9854\n");
    //if(PC == 0xAF36) BotFrontMsg("ConsiderNinja, called from %X\n", prevpc);
    if(PC == 0xAF31)
        switch(prevpc)
        {
            case 0xAF11: /*BotFrontMsg("No enemy generated, music is wrong\n");*/ break;
            case 0xAF1B: /*BotFrontMsg("No enemy, timer is bad\n");*/break;
            case 0xAF15: /*BotFrontMsg("No enemy generated, blocked by $85\n");*/ break;
            case 0xAF99: state.Goal=-1; BotFrontMsg("Not generating hokage, because the stage is not clear\n"); break;
            default: BotFrontMsg("Not generating enemy, called from %X\n", prevpc);
        }
    if(PC == 0xAFC4)
    {
        //BotFrontMsg("Generate Enemy %02X (frame=%02X)\n", A, RAM[0x21]);
        //if(A < 0x40) state.Goal = -1;
    }
    //if(PC == 0x93E6) BotFrontMsg("Considering whether to enable ninjas again\n");
    if(PC == 0x9A61) BotFrontMsg("Reselecting ninja timer (called from %X)\n", prevpc);
    if(PC == 0x9A72 || PC == 0xB45D
    || PC == 0x8333 || PC == 0x848B
    || PC == 0x97A6 || PC == 0x97AE) BotFrontMsg("@%X: Ninja Timer Set as %02X (called from %X)\n", PC,
        ARead[PC](PC)==0x86 ? X : A, prevpc);
    //if(PC == 0x9404) BotFrontMsg("Ninjas enabled again\n");
    if(PC == 0xB410)
    {
        BotFrontMsg("kill one.. of type %02X\n", A);
        if(A >= 0x40)
        {
            BotFrontMsg("** HOKAGE KILLED @ x=%02X, seen=%u, cball=%u\n",
                RAM[0x8C+X],
                RAM[0x4A], RAM[0x4B]);
                
            if(0
           // || (RAM[0x4B] < 3 && RAM[0x8C+X] < 0x30)
           // || (RAM[0x21]<0xC3 && RAM[0x4A] >= 10)
           // || (RAM[0x21]<0xE3 && RAM[0x4A] >= 11)*/
            || 1
              )
                state.Goal += 3; // KAGE KILLED
        }
        else
        {
            state.Goal += 1;
            if(RAM[0x2E] != 0x03) state.Goal = -1;
            //if(RAM[0x21] >= 0x9E) state.Goal = -1; // too late
            BotFrontMsg("goal=%d\n", state.Goal);
        }
        //state.Goal += 1;
    }
    if(PC == 0xAF6C) BotFrontMsg("Enemy %u witholds a crystal ball!\n", X);
    //if(PC == 0x90E5) BotFrontMsg("%X called from %X\n", PC, prevpc);
    /*if(PC == nmi
    && (unsigned char)RAM[0x84] <= 8) { state.Goal = 2; }
    */
    if(PC == 0xAC69 && A >= 0xF8)
    {
        if(state.Goal == 0)
        {
            //BotFrontMsg("Enemy %u disappears\n", X);
            //state.Goal = -1;
        }
        else
        {
            //BotFrontMsg("Enemy %u disappears, doesn't matter\n", X);
        }
    }
    
    /*
    // KAGE APPEARS
    if(RAM[0x89] >= 0x40 && RAM[0x89] <= 0x7F
    || RAM[0x8A] >= 0x40 && RAM[0x8A] <= 0x7F
    || RAM[0x8B] >= 0x40 && RAM[0x8B] <= 0x7F) state.Goal = 2;
    */
    
    //if(RAM[0x4B] < 3) state.Goal = 2; // Crystal ball found 
    
    /*if(RAM[0x31] & 0x80)
    {
        for(unsigned a=0; a<3; ++a)
            if(RAM[0x98+a] > 0) // this one died
                if(RAM[0x8C+a] > 0xC0) // died at right edge
                    state.Goal = 2;
    }*/
#endif

#if 0 // SMB MARIO
    if(RAM[0xCE] >= 200) state.Goal = -2;
    
    if(PC == 0xB511 || PC == 0xB518)
    {
        if(!BotFrontActive())FCEU_DispMessage("boing");
        //state.Goal = 2; // jump success!
    }
    if(PC == 0xB04C) // magic_jT9+2
    {
        //fprintf(stderr, "JT9 with A=%02X\n", A);
        switch(A)
        {
            case 9: // eat mushroom
                if(!BotFrontActive())FCEU_DispMessage("shrOOM!");
                state.Goal = 2;
                break;
            case 1: // auto climb
            case 2: // enter horiz pipe
            case 3: // enter vert pipe
            case 4: // slide flagpole down
            case 5: // victory 
            case 12: // eat flower
                //state.Entry.Set(A);
                break;
            case 10:// eat damage
            case 11:// eat death
                //state.Damage.Set(A);
                //fprintf(stderr, "Pit (a=%u)\n", A);
                //state.Pit=1;
                break;
            case 0: // regen level
            case 7: // start level
            {
                //state.Pit    = 0;
                //state.MarioY = 128;
                break;
            }
            case 8: // normal level activity
            {
            /*
                state.Active = 1;
                
                int MarioYPos = state.MarioY;
                
                if(!state.Pit && MarioYPos >= 0xF0 && MarioYPos <= 0xFF)
                {
                    state.Pit = -1; // ignore, it's screen top
                }
                else if(state.Pit < 0 && MarioYPos < 0x60)
                    state.Pit = 0;
                else if(!state.Pit && MarioYPos > 0xC0 && MarioYPos < 0xFF)
                {
                    state.Pit=1;
                    fprintf(stderr, "Pit (y=%u)\n", MarioYPos);
                }
            */
                break;
            }
        }
    }
#endif

#if 0 //ROCKMAN 2

    double RockX  =
        RAM[0x440]*256 + (unsigned short)(RAM[0x460]*256 + RAM[0x480])/ 256.0;
    if(PC ==0xC10B)
    {
        /*if(!BotFrontActive())*/FCEU_DispMessage("Death x=%g", RockX);
        state.Goal = -2;
    }

    // 90C9 = DoScrolling
    if(true
    && (PC == 0x9185)) // vert scrolling
    {
        //FCEU_DispMessage("VertScroll at PC=%04X A=%02X X=%02X Y=%02X\n", PC, A,X,Y);
        //state.Goal += 0;
    }
    if(true
    && (PC == 0x8BCA))
    {
        //FCEU_DispMessage("Hit screen bottom");
    }
    if(true
    && (PC == 0x90D2)) // horiz scrolling
    {
        FCEU_DispMessage("HorizScroll at PC=%04X A=%02X X=%02X Y=%02X\n", PC, A,X,Y);
        //state.Goal += 2;
    }
    if(true
    && (PC == 0x8F39)) // next-scrolling
    {
        //if(RAM[0x37] != 3)
        BotFrontMsg("ScrollNextRoom at PC=%04X A=%02X X=%02X Y=%02X -- $37=%u, Rockman Y=%.2f speed %.2f, item1 %02X\n",
            PC, A,X,Y, RAM[0x37],
            (unsigned short)(RAM[0x4A0]*256 + RAM[0x4C0])/256.0,
            (signed short)(RAM[0x640]*256 + RAM[0x660])/256.0,
            RAM[0xA4]);
        //if(RAM[0x37] & 1) state.Goal = -5; else state.Goal += 2;
        state.Goal += 2;
        // vert = bad, horiz = good
    }
    if(true
    && (PC == 0x8EDD)) // prev-scrolling
    {
        //if(RAM[0x37] != 3)
        if(!BotFrontActive())
        BotFrontMsg("ScrollPrevRoom at PC=%04X A=%02X X=%02X Y=%02X -- $37=%u, Rockman Y=%.2f speed %.2f, item1 %02X\n",
            PC, A,X,Y, RAM[0x37],
            (unsigned short)(RAM[0x4A0]*256 + RAM[0x4C0])/256.0,
            (signed short)(RAM[0x640]*256 + RAM[0x660])/256.0,
            RAM[0xA4]);
        state.Goal = -2;
    }
    if(PC == 0x80AB)
    {
        FCEU_DispMessage("Boss %u damaged with weapon %u (%d points left)\n", RAM[0xB3], RAM[0xB3], (signed char)A);
    }
    if(PC == 0xF262)
    {
        FCEU_DispMessage("Create no bonus\n");
    }
    if(PC == 0xF27B)
    {
        FCEU_DispMessage("Create bonus 0x%02X\n", A);
        //if(A == 0x78) state.Goal += 2; else state.Goal -= 5; 
    }
    
    if(!BotFrontActive())
    {
        if(/*PC == 0xCFFA // megaman2
        || */PC == 0xCFF7 // rockman2
          ) FCEU_DispMessage("LAG detected at PC=%04X\n", PC);
    }
#if 0
    if(RAM[0x6C1] < 28 - 1*6)
    {
        state.Goal += 2;
        if(BotFrontActive())
            FCEU_DispMessage("Boss HP is %u\n", RAM[0x6C1]);
    }
#endif
    
    /*
    static int counter=0;
    if(++counter>=50000) {counter=0;
    BotFrontMsg("Object types: ");
    for(unsigned a=0; a<32; ++a)
        BotFrontMsg(" %u", RAM[0x400 + a]);
    BotFrontMsg("\n");}*/
    
    /*if(PC == nmi)
    {
        FCEU_DispMessage("Y=%.2f speed=%.2f\n",
            (unsigned short)(RAM[0x4A0]*256 + RAM[0x4C0])/256.0,
            (signed short)(RAM[0x640]*256 + RAM[0x660])/256.0);
    }*/
#endif

#if 0 //ROCKMAN 3
    double RockY  = (unsigned short)(RAM[0x3C0]*256 + RAM[0x3A0])/ 256.0;
    double RockX  =
        RAM[0x380]*256 + (unsigned short)(RAM[0x360]*256 + RAM[0x340])/ 256.0;
    double RockYS = (signed short)(RAM[0x460]*256 + RAM[0x440])/ 256.0;
#endif

#if 0
    if(false && PC == 0x92B6)
    {
        FCEU_DispMessage("stage=%02X A=%02X CurScreen=%02X\n", RAM[0x31], A, RAM[0x460]);
    }
    if(false && PC == 0xCD8A)
    {
        FCEU_DispMessage("Y=%u 0C=$%02X 0D=$%02X 0E=$%02X\n", Y,RAM[12],RAM[13],RAM[14]);
        BotFrontMsg("#Actives=%u [%f] (", RAM[0x720], RAM[0x8E]/6.0);
        for(unsigned a=0; a<RAM[0x720]; ++a)
        {
            BotFrontMsg("%u,%u,%u,%u,%u,%u, ",
                RAM[0x721+a*6],
                RAM[0x722+a*6],
                RAM[0x723+a*6],
                RAM[0x724+a*6],
                RAM[0x725+a*6],
                RAM[0x726+a*6]);
        }
        BotFrontMsg(")\n");
    }
    //if(PC == 0xCDDE) BotFrontMsg("CDDE y=%u\n", Y);
#endif
}

void BisqBotReset()
{
    struct timeval tv;
    gettimeofday(&tv,NULL);
    rng.seed(tv.tv_usec + tv.tv_sec*1000000ULL);
    
    state.Reset();
}

unsigned char BisqBotInput_Do(unsigned FrameNo);
unsigned char BisqBotInput(unsigned FrameNo)
{
    unsigned char res = BisqBotInput_Do(FrameNo);
    static std::vector<unsigned char> BestInput;
    if(BestInput.empty())
    {
    #define d(c,k) for(size_t n=0; n<c; ++n)BestInput.push_back(k);
//d(2,K_A) d(1,0) d(1,K_A) d(2,0) d(1,K_A) d(1,0) d(1,K_A) d(1,0) d(5,K_A) d(3,0) d(3,K_A) d(2,0) d(1,K_A) d(3,0) d(1,K_A) d(1,0) d(1,K_A) d(1,0) d(6,K_A) d(21,0) d(1,K_A) d(4,0) d(1,K_A) d(1,0) d(1,K_A) d(4,0) d(1,K_A) d(4,0) d(2,K_A) d(2,0) d(1,K_A) d(1,0) d(1,K_A) d(6,0) d(2,K_A) d(1,0) d(2,K_A) d(1,0) d(2,K_A) d(4,0) d(1,K_A) d(1,0) d(4,K_A) d(2,0) d(2,K_A) d(1,0) d(1,K_A) d(2,0) d(1,K_A) d(1,0) d(2,K_A) d(2,0) d(1,K_A) d(2,0) d(1,K_A) d(4,0) d(4,K_A) d(1,0) d(1,K_A) d(2,0) d(2,K_A) d(1,0) d(3,K_A) d(2,0) d(2,K_A) d(4,0) d(1,K_A) d(4,0) d(1,K_A) d(1,0) d(2,K_A) d(4,0) d(2,K_A) d(3,0) d(3,K_A) d(3,0) d(3,K_A) d(4,0) d(4,K_A) d(1,0) d(2,K_A) d(1,0) d(2,K_A) d(1,0) d(1,K_A) d(4,0) d(4,K_A) d(4,0) d(1,K_A) d(1,0) d(4,K_A) d(3,0) d(1,K_A) d(1,0) d(1,K_A) d(1,0) d(1,K_A) d(1,0) d(2,K_A) d(3,0) d(2,K_A) d(1,0) d(1,K_A) d(1,0) d(4,K_A) d(1,0) d(1,K_A) d(2,0) 
    #undef d
    }
    //if(FrameNo < 238) return BestInput[FrameNo];
    return res;
}
unsigned char BisqBotInput_Do(unsigned FrameNo)
{
#if 0 /* ROCKMAN 1 */
    double RockX  = (unsigned short)(RAM[0x480]*256 + RAM[0x4A0])/ 256.0 + RAM[0x460]*256;
    double RockY  = (unsigned short)(RAM[0x600]*256 + RAM[0x620])/ 256.0;
    double RockYS = (signed short  )(RAM[0x680]*256 + RAM[0x660])/ 256.0;
#endif
#if 1 /* ROCKMAN 2 */
    double RockY  = (unsigned short)(RAM[0x4A0]*256 + RAM[0x4C0])/ 256.0;
    double RockX  =
        RAM[0x440]*256 + (unsigned short)(RAM[0x460]*256 + RAM[0x480])/ 256.0;
    double RockYS = (signed short)(RAM[0x620]*256 + RAM[0x640])/ 256.0;
#endif
#if 0
    double NextY = RockY-RockYS;
#endif

#if 0 /* Rockman 2: Bubbleman downscroll */
    static unsigned ChangingWeapons,shot;
    if(FrameNo==0) ChangingWeapons=shot=0;
ReKey:
    if(ChangingWeapons) { if(ChangingWeapons++ == 15) return K_ST;
                          if(ChangingWeapons==43) {ChangingWeapons=0; goto ReKey; }
                          return K_R; }
    if(!shot && GetRnd<8>()==0) { ChangingWeapons=1; return K_ST; }
    return ((GetRnd<40>()!=0) ? K_R : K_L)
         | ((RockY >= 180 && GetRnd<8>()==0) ? (shot=K_B) : 0);
#endif

    if(RAM[0x37] >= 240
    && RAM[0x51] >= 240
    && RAM[0x6B] >= 240) { state.Goal=2; return 0; }
    return GetRnd<2>() ? K_A : 0;

#if 0 /* Rockman 2: Crashman zipper and downscroll */

    /* Program description:
     
        1. Idle for N frames, hanging in the ladder facing right.
        2. Fire an Item1 to the right.
        3. Idle for N frames. Meanwhile, Item1 elevates.
        4. Press up for 1 frame (but without waiting, go to step 5)
        5. Hold left for N frames, to enable falling down before zipping too far
        6. Fall until X >= 4296 (plus 0..2 frames)
        7. Hold A for 3 frames
            Hold Left so that X >= 4344 is not reached before Y <= 208
        8. Hold left so that X >= 4456 is not reached before Y >= 208
            Meanwhile, shoot if you want
        9. Press left sometimes if you want
           Shoot sometimes if you want
           
     
    */
    //if(FrameNo == 0 && RockX >= 4344) { state.Goal=-2; return 0; }

    static unsigned Idle1Length, Idle2Length, Zip0hold,Zip0wait;
    static unsigned JumpLength,JumpLeft,JumpHoldA;
    static unsigned IgFrame;
    if(FrameNo==0) {
        Idle1Length = GetRnd<70>();
        Idle2Length = GetRnd<8>();
        Zip0hold = 3+GetRnd<5>();
        Zip0wait = 4296-16 + GetRnd<3>()*16; IgFrame=0;
        JumpLeft = 3 + GetRnd<2>();
        //JumpLeft=3; 
        JumpHoldA=3+GetRnd<2>();
        JumpLength=std::max(JumpHoldA,JumpLeft);
        
        //JumpLeft=4; JumpHoldA=3;
    }
    if(FrameNo < Idle1Length)
    {
        if(GetRnd<20>() == 0) return K_B|K_R;
        return 0;
    }
    FrameNo -= Idle1Length; // step 1
    if(FrameNo == 0) { return K_B; }   FrameNo -= 1;        // step 2
    if(FrameNo < Idle2Length) return 0; FrameNo -= Idle2Length; // step 3
    if(FrameNo < Zip0hold) return (K_L | (FrameNo==0 ? K_U : 0));
                                        FrameNo -= Zip0hold;    // step 4..5
    if(RockX < Zip0wait) { IgFrame=FrameNo+1; return 0; }
                                        FrameNo-=IgFrame;       // step 6
    //if(RockX >= 4360 && RockX <= 4400 && RockY >= 208) { state.Goal=-2; return 0; }
    if(FrameNo < JumpLength)
    {
        if(FrameNo == 1 && RockYS > 2) { state.Goal=-2; return 0; }
        return (FrameNo<JumpLeft)*K_L + (FrameNo<JumpHoldA)*K_A;
    }
                                        FrameNo-=JumpLength;    // step 7
    //if(FrameNo == 0 && RockYS < -0.75) { state.Goal=-2; return 0; }
    
    if(FrameNo >= 5 && RockX < 4344.0+0)  { state.Goal=-2; return 0; }
    if(FrameNo >= 7 && RockX < 4344.0+16) { state.Goal=-2; return 0; }
    if(RockX == 4344.0 && RockY > 214) { state.Goal=-2; return 0; }
    double DistanceTo4456 = (4456.0 - RockX) / 16.0;
    if(DistanceTo4456 > 0)
    {
        double FallingSpeed = -RockYS;
        int N_left_required = 0;
        for(;;)
        {
            int Nframes = (DistanceTo4456+0.5) + N_left_required;
            double y = RockY, v = FallingSpeed;
            while(Nframes > 0) { y += v; v += 0.25; --Nframes; }
            if(y >= 208) break;
            ++N_left_required;
        }
        unsigned left = 0;
        if(N_left_required > 0)
            if(GetRnd<2>() == 0 || DistanceTo4456 < 2)
                left = K_L;
        return left | (GetRnd<20>() == 0 ? K_B : 0);
    }
    if(RockX >= 4456 && RockY < 208) { state.Goal=-2; return 0; }

    unsigned char rt = (RockX >= 4536) ? K_R
                     : (GetRnd<6>()==0 ? K_L : 0);
    unsigned char a = 0;//K_A*(GetRnd<20>()==0);
    if(RockX >= 4500 && GetRnd<3>() == 0) return K_B|a|rt;
    //static bool fright=false;
    //if(fright || GetRnd<20>() == 0) {fright=false;return K_L|a;}
    //if(GetRnd<10>() == 0) {fright=true;return K_R|a;}
    return rt;
#endif

#if 0
    //if(FrameNo >= 20) return K_R;
    return (GetRnd<5>()==0 ? K_ST : 0);
#endif
    
    static const unsigned char InputTable[] =
    {
        K_R | K_B | K_A,
        K_R | K_B,
        K_R,
        K_L | K_B | K_A,
        K_L | K_B,
        K_L,
        K_L | K_R | K_B | K_A,
        K_L | K_R | K_B,
        K_L | K_R,
        K_A,
        0
    };
#if 0
    unsigned char res = InputTable[ GetRnd<sizeof(InputTable)>() ];
    if(GetRnd<3>()==0) res |= K_A;
    return res;
#endif
    
    static unsigned LastInput = 0;
    static unsigned A_State   = 0;
    static unsigned B_State   = 0;
    

    //if(FrameNo >= 26+3) return K_R;
   
    //if(LastInput == 0 && FrameNo < 50) LastInput |= K_A | K_R;
    //if(LastInput & K_L && FrameNo < 50) { LastInput &= ~K_L; LastInput |= K_R; }
    
    unsigned PrevLastInput = LastInput;
    if(FrameNo == 0) { PrevLastInput = LastInput = K_R|K_A|K_B; A_State = B_State = 0; }

#if 0 /* ROCKMAN 1 BEAMS -- Find the one that is below */
    if(LastInput & K_U)
    {
        for(unsigned a=0; a<16; ++a) { if(!RAM[0x5B0+a]) continue;
            if(RAM[0x5F0+a] >= RockY+14 && RAM[0x5F0+a] < RockY+16) NextChange=0;
        }
    }
    else
    {
        for(unsigned a=0; a<16; ++a) { if(!RAM[0x5B0+a]) continue;
            if(RAM[0x5F0+a] >= RockY+15 && RAM[0x5F0+a] < RockY+17) NextChange=0;
        }
    }
#endif
    
    if(!NextChange) 
    {
        LastInput = InputTable[ GetRnd<sizeof(InputTable)>() ];
        //if(GetRnd<40>() == 0) LastInput |= K_D;
        
      /*
        LastInput = 0;
        if(GetRnd<10>() == 0) LastInput |= K_A;
        if(GetRnd<5>() == 0) LastInput |= K_B;
        switch(GetRnd<20>())
        {
            //case 0: LastInput |= K_U; break;
            case 1: LastInput |= K_D; break;
        }
        switch(GetRnd<5>())
        {
            case 0: case 1: LastInput |= K_L; break;
            case 2: case 3: LastInput |= K_R; break;
        }
      */
        
        //if(LastInput & K_U)
        //    NextChange = GetRnd<4>();
        //else
            NextChange = GetRnd<8>();
    }
    else --NextChange;
    
#if 0 /* KAGE */
   { unsigned char In = LastInput;
    //if(FrameNo >= 20 && GetRnd<30>() == 0) In = (In ^ (K_L | K_R)) | K_B;
    
    /*if(state.Goal == 0) {
    static bool LastStop=false;
    if(RAM[0x8F] < 0xF0
    && RAM[0x8C] >= 0xF5 && !LastStop) { In &= ~(K_L|K_R); LastStop=true; }
    else LastStop=false;

    }*/
    //if(FrameNo < 90) In &= ~K_B;
    
    return In;}
#endif

    //if(RockX >= 6757 && RockX < 6768) LastInput |= K_A;
    
    //if(RockX >= 6740 && RockX < 6756) LastInput &= ~K_A;
    //if(FrameNo < 15) { LastInput &= ~K_L; LastInput |= K_R; }
    //if(FrameNo < 15) { LastInput = K_R | K_A; }
/*
    if(RockY >= 1.5) LastInput = K_R;
*/
    // if(RockX + (26+3-FrameNo) * 1.375 < 6287.0) { state.Goal = -1; }
    
    /*
    if(B_State && FrameNo)
    {
        double RockX = (unsigned short)(RAM[0x480]*256 + RAM[0x4A0])/ 256.0;
        fprintf(stderr,
            "[frame %u]I think the beam went at %u... X=%.1f screen=%u\n",
                FrameNo, RAM[0x5F5], RockX, RAM[0x460]);
    }
    
    B_State = (NextY >= 112 && NextY <= 114) ? K_B : 0;
    if(B_State) BotFrontMsg("Try beam at RockY=%.1f (%+.1f)\n", RockY, RockYS);
    if(RockY >= 104 && RockY <= 112 && NextY > 113)
        BotFrontMsg("Rats, this try failed Y=%.1f (%+.1f)\n", RockY, RockYS);
    */
    
    /*
    if(FrameNo < 7) if(GetRnd<3>() == 0) Input |= K_SE;
    if(FrameNo >= 12 && FrameNo < 45) if(GetRnd<20>() == 0) Input |= K_ST;
    */
    
    //if(FrameNo < 16) A_State = K_A; else A_State = 0;
    //if(FrameNo < state.EventFrame) A_State = K_A; else A_State = 0;
    //if(GetRnd<10>() == 0) A_State ^= K_A;
    //A_State = K_A;
    //if(FrameNo < 43) B_State = K_B; else B_State = 0;

#if 0
    bool Region = false; //RockY < 88.0;
    if(state.NumShoots < state.MaxShoots
        /*
       && !(LastInput & K_U)
       && state.EventFrame == 0
        */
       && B_State == 0
       && ((Region && GetRnd<20>() == 0)
       || (!Region && GetRnd<35>() == 0))
      )
        { ++state.NumShoots; B_State = K_A | K_B;
          //if(LastInput & K_L) LastInput ^= (K_R | K_L);
          if(Region) state.Region=true;
          /*
          state.EventFrame = 15; //FrameNo;
          */
        }
    else
        B_State = 0;
#endif

    /*
    if(state.EventFrame) --state.EventFrame;
    
    if(!(LastInput & K_U))
    {
        if(RockYS >= 0.3 && RockYS <= 0.63)
        {
            LastInput = K_U;
            if(RockX < 9425) state.Goal = -1;
        }
        if(RockYS >= 0 && RockYS <= 0.3) state.Goal = -1;
    }
    */
    
    /*
    if(state.EventFrame>0) --state.EventFrame;
    if(state.EventFrame==1) { state.EventFrame = 0; B_State = K_B; }
    else { B_State = 0; }

    if(RockX >= 6388.0 && RockX <= 6391.0) state.MaxShoots = 14;
    if(state.MaxShoots > 0) { A_State=K_A; --state.MaxShoots; } else A_State=0;
    if(LastInput && !PrevLastInput) A_State = K_A;
    */
   
    //if(B_State != 0 && GetRnd<20>() == 0) LastInput |= K_U;
    
    //if(B_State == K_B) { LastInput &= ~K_L; LastInput |= K_R; }
    //if(FrameNo >= 8 && FrameNo < 15) LastInput |= K_A;
    
    //if(GetRnd<15>() == 0) A_State = K_A; else A_State = 0;
    //if(GetRnd<15>() != 0) A_State = K_B; else B_State = 0;
    //if(GetRnd<20>() == 0) Input |= K_U;
    //if(GetRnd<10>() == 0) Input |= K_ST;
    //if(GetRnd<100>() == 0) Input |= K_SE;
    //if(RockY < 1) A_State = 0;
    
    //if((PrevLastInput & K_U) && (LastInput & K_A)) NextChange=0;

    unsigned char Input = 0x00;
    Input |= LastInput;
    Input |= B_State;
    Input |= A_State;
    
#if 1 // SMB1 pipe entry
    Input |= K_D;
#endif

    //if(GetRnd<50>() == 0) Input ^= K_R | K_L;
    //else if((Input & K_A) && GetRnd<50>() == 0) Input &= ~(K_R | K_L);
    
    return Input;
}

extern "C" {
#include <inttypes.h>
extern int current; extern uint32_t framecount; }
void BisqBotRAMreport()
{
   char msg[16384*4] = "";
   extern unsigned char RAM[0x800];
#if 0 /* LUNAR BALL */
   sprintf(msg, "Pow[%d]Ang[%d]",
     RAM[0x3A0],
     RAM[0x3C0]
   );
   
#endif
#if 0 /* ROCKMAN 1 */
   double RockY  = (unsigned short)(RAM[0x600]*256 + RAM[0x620])/ 256.0;
   double RockYS = (signed short  )(RAM[0x680]*256 + RAM[0x660])/ 256.0;
   double RockX  = (unsigned short)(RAM[0x480]*256 + RAM[0x4A0])/ 256.0 + RAM[0x460]*256;
   double RockXS = (signed short  )(RAM[0x4C0]*256 + RAM[0x4E0])/ 256.0;
   
/*
   sprintf(msg, "\nPosi=(%.4f,%.4f)\nVelo=(%.4f,%.4f)",
    RockX,RockY, RockXS,RockYS);
*/

   sprintf(msg, " - X=%.2f%+.4f Y=%.2f%+.4f",
    RockX,RockXS, RockY,RockYS);
    
   if(1)
   {
     char mdata[512];
     sprintf(mdata, "; random=%02X[%02X,%02X]", RAM[0x46],RAM[0x23],RAM[0x0D]);
     strcat(msg, mdata);
   }
    
   if(1)
   {
    unsigned char seed=RAM[0x46], frame=RAM[0x23], a;
    #define Iterate(n, c) \
      for(a=0;a<n;++a) { seed ^= c; seed += ++frame; seed >>= 1; }
    Iterate(3, 0x64);
    Iterate(3, 0x7C);
    Iterate(6, 0x64);
    Iterate(1, 0x54);
    Iterate(32, 0x03);
    Iterate(60, 0x02);
    Iterate(1, 0x54); 
    Iterate(33, 0x03);
    Iterate(2, 0x64); 
    Iterate(3, 0x7C); 
    Iterate(6, 0x64); 
    Iterate(1, 0x98); 
    Iterate(13, 0xD4);
    Iterate(1, 0xCC); 
    Iterate(81, 0xD0);
    {char mdata[512];
     sprintf(mdata,",Bomb[%02X],Boss[Strat=%u,Delay=%u]", seed,
        RAM[0x3E],RAM[0x641]);strcat(msg, mdata);}
    
    /* Bombman value needs to be EVEN at teleport entry, for Bombman to do a tiny jump.
     * However, the teleport collision check is only done on 1/2 of frames.
     * See randsim.cc
     */
   }
   
   if(0 && RockYS <= 0.0)
   {
     strcat(msg, "; jmps=");
     if(RAM[0x620] != 0 && RockYS != -0.75)
     {
       double NewRockY = RockY - RockYS;
       /* Take fraction part only */
       RockY = ((int)RockY) + (NewRockY - (int)NewRockY);
     }
     {double JmpS = 4.8710938;
     unsigned n;for(n=0; n<20; ++n)
     {
         char mdata[512];
         sprintf(mdata, "%d,", (signed int)RockY);
         strcat(msg, mdata);
         RockY -= JmpS;
         JmpS -= 0.25;
         if(RockY < -4) break;
     }}
   }
   
   {unsigned a,b=0;for(a=0; a<16; ++a)
   {
     char mdata[512];
     if(RAM[0x5B0+a]==0) continue;
     
     if(0)
       { if(b==0) { ++b; strcat(msg, "\n[36mBeams: "); } }
     else
       { if(b==0) { ++b; strcat(msg, "\nBeams: "); } }
     
     unsigned pixlen = (RAM[0x5A0]-1)*2+16; if((pixlen&7)==0)pixlen+=8;
     sprintf(mdata, "%u,%u+%u[%u,%u], ",
        RAM[0x5D0+a]*256 + RAM[0x5C0+a],
        RAM[0x5F0+a],
        RAM[0x5A0+a],
        RAM[0x5E0+a], RAM[0x5B0+a]
        //RAM[0x5D0+a]*256 + RAM[0x5C0+a] + pixlen
        );
   
     strcat(msg, mdata);
   }}
   if(0) strcat(msg, "[37m");
   
   {unsigned a,b=0;for(a=0; a<32; ++a)
   {
     char mdata[512];
     
     if(RAM[0x600+a] == 0xF8) continue; // dead object
     if(RAM[0x6E0+a] != 0x2C) continue; // lift animator
     
     double LiftX, LiftY;
     if(b==0)
     {
        double LiftX  = (signed short)(RAM[0x9D]*256 + RAM[0x9C])/ 256.0;
        ++b;
        sprintf(mdata, "\n[33mLifts (9B=%02X,9E=%02X,Speed=%.1f): ",
            RAM[0x9B],
            RAM[0x9E],
            LiftX);
        strcat(msg, mdata);
     }

     LiftX  = (unsigned short)(RAM[0x480+a]*256 + RAM[0x4A0+a])/ 256.0 + RAM[0x460+a]*256;
     LiftY  = (unsigned short)(RAM[0x600+a]*256 + RAM[0x620+a])/ 256.0;

     sprintf(mdata, "%.0f,%.0f%s, ", LiftX,LiftY, (RAM[0x640+a]&64) ? "R" : "L");
   
     strcat(msg, mdata);
   }}
   
   if(0)
     strcat(msg, "[37m");

#endif
#if 0 /* CASTLEVANIA */
   double RockY  = RAM[0x3F];
   double RockX  = (unsigned short)(RAM[0x41]*256 + RAM[0x40]);
   sprintf(msg, " - X=%.2f Y=%.2f\n",
    RockX,RockY
     );
#endif
#if 0 /* ROCKMAN 2 */
   double RockY  = (unsigned short)(RAM[0x4A0]*256 + RAM[0x4C0])/ 256.0;
   double RockYS = (signed short  )(RAM[0x640]*256 + RAM[0x660])/ 256.0;
   double RockX  = (unsigned short)(RAM[0x460]*256 + RAM[0x480])/ 256.0 + RAM[0x440]*256;
   double RockXS = (signed short  )(RAM[0x600]*256 + RAM[0x620])/ 256.0;
   sprintf(msg, " - X=%.2f%+.3f Y=%.2f%+.3f\n",
    RockX,RockXS, RockY, RockYS
     );
   {
    /*
    sprintf(Buf, "6E0=[%u,%u,%u,%u, %u,%u,%u,%u, %u,%u,%u,%u, %u,%u,%u,%u]",
        RAM[0x6E0], RAM[0x6E1], RAM[0x6E2], RAM[0x6E3],
        RAM[0x6E4], RAM[0x6E5], RAM[0x6E6], RAM[0x6E7],
        RAM[0x6E8], RAM[0x6E9], RAM[0x6EA], RAM[0x6EB],
        RAM[0x6EC], RAM[0x6ED], RAM[0x6EE], RAM[0x6EF]);
    */
    char mdata[512];
    sprintf(mdata, " B0[%u %u %u], Boss[40:%u 42:%u,4E:%u,,68:%u 6A:%u 6C:%u 6E:%u Ypower=%d]",
        RAM[0xB0], RAM[0xB1], RAM[0xB2],
        RAM[0x401],
        RAM[0x421],
    //    RAM[0x441],
    //    RAM[0x461],
    
    //    RAM[0x481],
    //    RAM[0x4A1],
    //    RAM[0x4C1],
        RAM[0x4E1],
        
    //    RAM[0x601],
    //    RAM[0x621],
    //    RAM[0x641],
    //    RAM[0x661],
        
        RAM[0x681],
        RAM[0x6A1],
        RAM[0x6C1],
        RAM[0x6E1],
        
        RAM[0x641]
        );
      strcat(msg, mdata);
   }
#endif
#if 0 /* Battle of olympus */
   // 18 = buttons down this frame (A, B only?)
   // F8 = buttons down this frame
   // F9 = buttons down
   double RockX  = (unsigned short)(RAM[0x10]*256 + RAM[0x52])/256.0;
   double RockY  = (unsigned short)(RAM[0x12]*256 + RAM[0x57])/256.0;
   double RockXS = ((signed char)RAM[0x53]) / 16.0;
   sprintf(msg, " - X=%.2f%+.3f Y=%.2f%+.3f frame=%u\n",
    RockX, RockXS, RockY, 0.0,
    RAM[0x47]);
#endif
#if 0 /* KAGE */
   {unsigned frames_until_enemy = 0;
   for(unsigned n=0; n<256; ++n,++frames_until_enemy)
     if((RAM[0x57] & (unsigned char)(RAM[0x21] + n)) == 0) break;
   sprintf(msg, " - NextEnm@%u(%02X&%02X),SrNo=$%X,Kil=%u,Seen=%u(cball=%u),Remn=$%02X,Hokage=%u, Kage=%02X,%02X, Butterfly=%02X,%02X,6D2=%X EX=%X",
    frames_until_enemy, RAM[0x21],RAM[0x57],
    RAM[0x2E],//screen
    RAM[0x4D],//killed
    RAM[0x4A],//seen
    RAM[0x4B],//crystal ball
    RAM[0x84],//remain
    RAM[0x31],//hokageflag
    
    RAM[0x53],RAM[0x54],
    
    RAM[0xB7],RAM[0xB8],
    
    RAM[0x6D2],
    
    RAM[0x8C]
     );
   };
#endif
#if 0 /* HINOTORI */
   sprintf(msg, " - Hit=%u", RAM[0xD9]);
#endif
#if 0 /* ROCKMAN 4 */
   double RockY  = (unsigned short)(RAM[0x378]*256 + RAM[0x360])/ 256.0;
   double RockYS = (signed short  )(RAM[0x3F0]*256 + RAM[0x3D8])/ 256.0;
   double RockX  = (unsigned short)(RAM[0x330]*256 + RAM[0x318])/ 256.0 + RAM[0x348]*256;
   double RockXS = 0;
   sprintf(msg, " - X=%.2f%+.3f Y=%.2f%+.3f\n",
    RockX,RockXS, RockY, RockYS
     );
#endif
#if 0 /* ROCKMAN 3 */
   double RockY  = (unsigned short)(RAM[0x3C0]*256 + RAM[0x3A0])/ 256.0;
   double RockYS = (signed short  )(RAM[0x460]*256 + RAM[0x440])/ 256.0;
   double RockX  = (unsigned short)(RAM[0x360]*256 + RAM[0x340])/ 256.0 + RAM[0x380]*256;
   double RockXS = 0;
   sprintf(msg, " - X=%.2f%+.3f Y=%.2f%+.3f\n",
    RockX,RockXS, RockY, RockYS
     );
#endif
#if 1 /* SMB MARIO */
    double RockY  = RAM[0xCE];
    double RockX  = RAM[0x86];
   sprintf(msg, " - X=%.2f Y=%.2f\n",
    RockX, RockY
     );
#endif
#if 0 /* SMB3 */
    double RockY  = RAM[0x87]*256 + RAM[0xA2];
    double RockX  = RAM[0x75]*256 + RAM[0x90];
   sprintf(msg, " - X=%.2f Y=%.2f\n",
    RockX, RockY
     );
#endif
#if 0 /* GREMLINS 2 */
   #define Load3(addr) \
      ((signed int)((RAM[addr+0]<<24) | (RAM[addr+1]<<16) | (RAM[addr+2]<<8))) / 65536.0
   double RockX  = Load3(0x40);
   double RockXS = Load3(0xA3);
   double RockYS = Load3(0xA0);
   double RockY  = Load3(0x43);
   #undef Load3
   sprintf(msg, " - X=%.2f%+.3f Y=%.2f%+.3f\n",
    RockX,RockXS, RockY, RockYS
     );
#endif
#if DO_RAMDUMP
    /*
        Red = changed smaller
        Green = changed bigger
        Cyan = is not changing anymore
        White = preserved
    */
   {static unsigned char PrevRAM[0x800];
    static char PrevChanges[0x800] = {0};
   {unsigned a;for(a=0x300; a<0x400; a+=0x01)
   {
     //if(a==0x24)a=0x400;
     char mdata[512];
/*
     if(a != 0x553 && a != 0x554 && a != 0x473 && a != 0x474
     && a != 0x393 && a != 0x394 && a != 0x6F && a != 0x1A) continue;
*/
     
     //if(RAM[a] == PrevRAM[a]) continue;
     //if(RAM[a] < 0x4A || RAM[a] > 0x4E) continue;
     //if(a == 0x84) goto Ok;
     /*
     if(a < 0x66 && a != 0x2C && a != 0x2E
                 && a != 0x46 && a != 0x48
                 && a != 0x4C && a != 0x4D
                 && a != 0x56 && a != 0x2B
                 && (a < 0x30 || a > 0x35)) continue;
     if(a > 0x67 && a < 0x6D) continue;
     if(a > 0x6F && a < 0x78) continue;
     if(a > 0x7A && a < 0x86) continue;
     */
     if(
      /*a >= 0x62 && a <= 0x67 && !((a-0x62)%3)
     || a >= 0x6D && a <= 0x6F && !((a-0x6D)%3)
     || a >= 0x78 && a <= 0x7A && !((a-0x78)%3)
     || a >= 0x86 && a <= 0xB2 && !((a-0x86)%3)
     || ((a < 0x60 || a > 0xB2) &&*/ 
     /*
         (a >= 0x210 && a <= 0x2A0 && !((a-0x210)%0x08))
      || (a >= 0x4F7 && a <= 0x600 && !((a-0x4F7)%0x08))
      || (a >= 0x000 && a <= 0x01D && !((a-0x000)%0x08))
      || (a >= 0x01D && a <= 0x100 && !((a-0x01D)%26))
      */
        !(a%16)
       )
     {
     Ok:
       sprintf(mdata, "[%04X]", a);
       strcat(msg, mdata);
     }
     
     std::string color;
     if(PrevChanges[a])
     {
         if(RAM[a] < PrevRAM[a])      { color=("\33[31m"); } // red
         else if(RAM[a] > PrevRAM[a]) { color=("\33[32m"); } // green
         else color=("\33[36m"); // cyan
     }
     else
     {
         if(RAM[a] < PrevRAM[a])      { color=("\33[1;31m"); } // bright red
         else if(RAM[a] > PrevRAM[a]) { color=("\33[1;32m"); } // bright green
         else {} // nothing
     }
     
     sprintf(mdata, "%s%02X ", color.c_str(), RAM[a]);
     strcat(msg, mdata);
     
     if(!color.empty()) strcat(msg, "\33[0m");
     PrevChanges[a] = RAM[a] != PrevRAM[a];
     PrevRAM[a]=RAM[a];
   }}
   }
#endif
   FCEU_DispMessage("%s #%u%s",(current<0) ? "Play" : "Rec", framecount, msg);
}

unsigned long long BisqBotEvaluate(unsigned FrameNo)
{
    /*
    #define OkBeam(y) ((y)==113)
    
    state.Goal=1;
    if(OkBeam(RAM[0x5F5]) || OkBeam(RAM[0x5F6])
    || OkBeam(RAM[0x5F7]) || OkBeam(RAM[0x5F8])
    || OkBeam(RAM[0x5F9]))
    {
        if(RAM[0x460] == 5) // screen 5
            state.Goal = 2;
    }*/
    
    /*
    // Fail if screen 5 isn't reached in time
    if(FrameNo >= 40 && RAM[0x460] != 5) { state.Goal=-1; return 0; }
    // Fail if Rockman falls off the beam too soon
    if(FrameNo <= 32 && (signed char)RAM[0x600] > 5) { state.Goal=-1; return 0; }
    if(state.Goal < 0) return 0;
    */
    //if(FrameNo==60) state.Goal = 2;
    
#if 0
    if(BotFrontActive())
    {
        // Fails to fill HP meter?
        if(FrameNo >= 153+3 && RAM[0x6C1] < 1)  {
            BotFrontMsg("rej, hp 0\n"); state.Goal = -1;
    DBG:
            BisqBotRAMreport();
            return 0; }
        if(RAM[0x6C1] == 28 && RAM[0x641] > 3) {
            BotFrontMsg("rej, boss acted wrong\n"); state.Goal = -1; goto DBG; }
        if(FrameNo >= 82+3 && RAM[0x440] != 0x19) {
            BotFrontMsg("rej, not reached room\n"); state.Goal = -1; goto DBG; }
        if(RAM[0x6C1] == 28
        && RAM[0xB1] == 2)
        {
            // Must have at least two frames with B0[3 3 0]
            BotFrontMsg("1 point\n");
            ++state.Goal;
            
            if(state.Goal == 1) goto DBG;
        }
    }
#endif

#if 0 /* ROCKMAN 1 */
    double RockY  = (unsigned short)(RAM[0x600]*256 + RAM[0x620])/ 256.0;
    double RockX  =
        RAM[0x460]*256 + (unsigned short)(RAM[0x480]*256 + RAM[0x4A0])/ 256.0;
    double RockYS = (signed short)(RAM[0x680]*256 + RAM[0x660])/ 256.0;
#endif
#if 0 /* ROCKMAN 2 */
    double RockY  = (unsigned short)(RAM[0x4A0]*256 + RAM[0x4C0])/ 256.0;
    double RockX  =
        RAM[0x440]*256 + (unsigned short)(RAM[0x460]*256 + RAM[0x480])/ 256.0;
    double RockYS = (signed short)(RAM[0x620]*256 + RAM[0x640])/ 256.0;
#endif
#if 0 /* SMB MARIO */
    double RockY  = RAM[0xCE];
    double RockX  = RAM[0x86];
    double RockYS = 0; // unknown
    //if(RAM[0x86] == 99 && !(RAM[0xCE] % 16)) state.Goal = 2;
    if(RockX > 0x8C && RockX < 0xB0 && RockY > 0x80)
    {
        state.Goal = 2; // done
    }
    
    if(false && RockX < 14 && RockY > 16)
    {
        if((int)RockY % 16) state.Goal = -2;
        else
        {
            if(!BotFrontActive())FCEU_DispMessage("boing?");
            state.Goal = 2; // jump possible
        }
    }
#endif
#if 0 /* SMB3 */
    double RockY  = RAM[0x87]*256 + RAM[0xA2];
    double RockX  = RAM[0x75]*256 + RAM[0x90];
    double RockYS = 0; // unknown
    if(RockX >= 116+8
    && RockY >= 320+8
    )
    {
        fprintf(stderr, "goal with %g, %g\n", RockX, RockY);
        state.Goal = 2;
    }
    if(RockX >= 116+4
    && RockY <= 320)   state.Goal = -2;
#endif
#if 0 /* CASTLEVANIA */
    double RockY  = RAM[0x3F];
    double RockX  = (unsigned short)(RAM[0x41]*256 + RAM[0x40]);
    double RockYS = 0;
#endif
#if 0 /* KAGE */
    //RockY = RAM[0x84]; /* # of enemies remaining */
    RockY = RAM[0x56] + RAM[0x2E]*256;
    
    /*
    if(RAM[0x21] >= 0x20  && RAM[0x21] < 0x60
    && state.Goal == 0) state.Goal = -1;
    */
    
    /*
    if(state.Goal == 1 && !(RAM[0x31] & 0x80))
    {
        // enemy killed but hokage not enabled = bad
        state.Goal = -1;
    }

    if(state.Goal == 0 && RAM[0x8F] < 0xF0 && RAM[0x8C] < 0x80) state.Goal = -1;
    if(state.Goal == 0 && RAM[0x8F] < 0xF0 && RAM[0x8F] < 0x80) state.Goal = -1;
    */
    
    /*
    if(state.Goal == 1
    && (!(RAM[0x31] & 0x80) // hokage was enabled?
     || RAM[0x21] >= 0x20 // it was too late
      )) state.Goal = -1;
    */
    
    //if(FrameNo >= 20 && state.Goal < 1) state.Goal = -1;
    
    //if(GoalFrameNo >= 36 && !(RAM[0x31] & 0x80)) state.Goal = -1;
    
    //if(RAM[0x84] == 0xF8 && FrameNo >= 200) state.Goal = -1;
#endif

    //if(state.Goal < 1 && RockY >= 117.0) { state.Goal = -5; }
    /**/
    //if(state.Goal == 0 && RockX >= 4456 && RockX < 4480 && RockY <= 196) state.Goal = -5;
    //if(state.Goal == 0 && RockX < 4520-48 && RockY >= 233) state.Goal = -5;
    /**/
   // if(RockX >= 1288) state.Goal = -5;
   
   //if(state.Goal == 0 && RockX >= 4210 && FrameNo==39) state.Goal = 2;
   //if(RockX < 4200) state.Goal = -1;
   
#if 0
    //if(state.Goal == 0 && RockY < 150) state.Goal=1;
   
    if(//RockX < 6410.0 || RockY >= 178.0 || RockYS <= -1.5
       RockY >= 100.0
    || (RockX >= 6759 && RockX < 6765 && RockYS < 0)
      )
    {
        state.Goal = -1;
    }
#endif

    if(state.Goal > 0)
    {
#if 0 /* Shot distances in Rockman 1 */
        double NearestDistance = 256;
        double NearestShot = RockX + 256;
        
        for(unsigned n=2; n<5; ++n)
        {
            if(RAM[0x600+n] == 0xF8) continue; // 
            double ShotX  =
                RAM[0x460+n]*256 + (unsigned short)(RAM[0x480+n]*256 + RAM[0x4A0+n])/ 256.0;
            double Distance = ShotX - RockX;
            if(Distance < 0) Distance = -Distance;
            
            if(Distance < NearestDistance)
                { NearestDistance = Distance; NearestShot = ShotX; }
        }
#endif

        // At walking speed, a pixel is worth 1/1.375 frames.
        
        //double rm1xscore  = 256*((RockX - 4500) / 1.1375);
        //double rm1xscore = 1000*(RockY / 0.75 - FrameNo);
        
        return 2000000000
            - FrameNo*256
         //   + (RockX-4200) * 256*256
         //   - (RockY)*256
            
         //     + rm1xscore
         //   + (unsigned)((RockX-6700)      *256)
         //   + (unsigned)((RockY*256*130/3))
         //   - state.MaxShoots*100
         //   - state.EventFrame*256*10
         //   + (unsigned)((state.BonusX-6700)*10000)
         //   + (unsigned)((NearestShot-6700)*256*30)
         ;
        
        //return (10000-FrameNo*3) + RAM[0xA4]*20;
        /*
        double RockY  = (unsigned short)(RAM[0x4A0]*256 + RAM[0x4C0])/ 256.0;
        double YSpeed = (  signed short)(RAM[0x640]*256 + RAM[0x660])/-256.0;
        
        return (10000-FrameNo*3) + (unsigned)(RockY + YSpeed*3); // ok
        */
    }

    return 0;//((RAM[0x460]-4)*256+RAM[0x480]); // not done
}

enum BisqBotInstruction BisqBotTick(unsigned FrameNo, unsigned Step)
{
    if(state.Goal < 0)
    {
        //fprintf(stderr,"rest %u\n", FrameNo);
        return BisqBotRestartStep;
    }
#if 0
    if(state.Goal < 2 && FrameNo >= 33 && !state.Region) return BisqBotRestart;
#endif
    
#if 0 /* For calculating Rockman axe throwing guys */
    //const unsigned reach_frame = 2855, reach_x = 0x79000;
    //const unsigned reach_frame = 2995, reach_x = 0x84000;
    //const unsigned reach_frame = 3115, reach_x = 0x8D400;
    const unsigned reach_frame = 3277, reach_x = 0x9A000;
    unsigned RockX  = RAM[0x480]*256 + RAM[0x4A0] + RAM[0x460]*256*256;
    unsigned frames_left = reach_frame - framecount;
    if(!frames_left)
    {
        unsigned RockXS = RAM[0x4C0]*256 + RAM[0x4E0];
        fprintf(stderr, "%g,%g?\n", RockX/256.0, RockXS*1.0);
        return RockX == reach_x+0x150 && RockXS == 0x160
                ? BisqBotAccept : BisqBotRestart;
    }
    else if(FrameNo > 1)
    {
        unsigned RockXS = RAM[0x4C0]*256 + RAM[0x4E0];
        if(RockXS == 0x160 && RockX + frames_left * 0x160 == reach_x+0x150)
            return BisqBotAccept;
        
        if(RockX + frames_left * 0x160 < reach_x)
        {/*
            fprintf(stderr, "restart @ %u (rem %u) (would reach only %g)\n",
                FrameNo,
                frames_left,
                (RockX + frames_left * 256*1.375) / 256.0);*/
            return BisqBotRestart;
        }
        if(RockX >= reach_x+0x100) return BisqBotRestart;
    }
    if(RAM[0x600] > 116 || RAM[0x600] < 100) return BisqBotRestart;
#endif

#if 0
    if(Step == 0 && state.Goal == 1)
    {
        state.MaxShoots = state.NumShoots+1;
        return BisqBotNextStep;
    }
#endif
    if(state.Goal >= 2)
    /*if((signed char)RAM[0x84]
       < (signed char)(1-Step))*/
    //if(RAM[0x31] < 0x80 && RAM[0x31] > Step)
    {
#if 0
            BotFrontMsg("** HOKAGE KILLED @ x=$%02X ($%02X @ screen, kage%+d),y=$%02X seen=%u, cball=%u, TIMER=%02X\n",
                (RAM[0x8C]-RAM[0x56]) + RAM[0x2E]*256,
                RAM[0x8C],
                RAM[0x8C]-RAM[0x53],
                RAM[0x8F],
                RAM[0x4A], RAM[0x4B],
                RAM[0x21]);
#endif

        //double RockY  = (unsigned short)(RAM[0x4A0]*256 + RAM[0x4C0])/ 256.0;
        //BotFrontMsg("RockY=%.2f, BonusX=%.3f, Goal=2\n", RockY, state.BonusX);
        if(Step+1 < MAX_PARTS) return BisqBotNextStep;
        return BisqBotAccept;
    }
    
    // Max item1s used: 5
    //if(RAM[0xA4] < 0x1C - 2*5) return BisqBotRestart;
    
    return BisqBotContinue;
}
